"""Implementation of Euler Angles."""

import numpy as np

import attlesnake.parameter.base as base
import attlesnake.parameter.constants as constants
import attlesnake.parameter.direction_cosine_matrix as direction_cosine_matrix


class BaseEulerAngle(base.BaseAttitudeParameter):
    """
    Base class for Euler Angle parameterizations.
    All angles are handled in radians by default except in
    methods explicitly denoted as working with degrees.
    """

    def __init__(
        self, a1: float = None, a2: float = None, a3: float = None
    ) -> None:
        """
        Initialize the Euler Angles based on the three angles (rad)
        about the sequence of axes (e.g. 3-2-1, 3-1-3, or others).
        """
        self.initialized = False
        self.a1 = self.a2 = self.a3 = None
        if all([a is not None for a in (a1, a2, a3)]):
            self.a1 = a1
            self.a2 = a2
            self.a3 = a3
            self.initialized = True

    def __repr__(self):
        return f"{self.a1:.4f}, {self.a2:.4f}, {self.a3:.4f}"

    @classmethod
    def from_dcm(
        cls, dcm: "direction_cosine_matrix.DCM"
    ) -> "BaseEulerAngle":
        """Initialize the Euler Angles instance from a DCM."""
        raise NotImplementedError("Derived classes should define this method")

    @property
    def rate_matrix(self) -> np.ndarray:
        """
        The rate matrix for the differential kinematic equation
        d(EA_321)/dt = rate_matrix @ angular_velocity_vector.
        """
        raise NotImplementedError("Derived classes should define this property")

    @property
    def array(self):
        return np.array([self.a1, self.a2, self.a3])

    @property
    def is_singular(self):
        """Whether the set of values is singular."""
        raise NotImplementedError("Derived classes should define this property")

    def print_rad(self):
        print(self)

    def print_deg(self):
        _a1, _a2, _a3 = self.array
        self.a1 = np.rad2deg(self.a1)
        self.a2 = np.rad2deg(self.a2)
        self.a3 = np.rad2deg(self.a3)
        print(self)
        self.a1 = _a1
        self.a2 = _a2
        self.a3 = _a3


class BaseSymmetricEulerAngle(BaseEulerAngle):
    @property
    def is_singular(self):
        """
        Symmetric Euler Angles are singular when the second angle
        is 0 or pi radians.
        """
        return (
            abs(self.a2) < constants.SINGULARITY_EPS
            or abs(self.a2 - constants.PI) < constants.SINGULARITY_EPS
        )


class BaseAsymmetricEulerAngle(BaseEulerAngle):
    @property
    def is_singular(self):
        """
        Asymmetric Euler Angles are singular when the second angle
        is +/- pi/2 radians.
        """
        return abs(abs(self.a2) - constants.PI/2) < constants.SINGULARITY_EPS


class EulerAngle321(BaseAsymmetricEulerAngle):
    """3-2-1 Euler Angle (yaw, pitch, roll)."""

    @property
    def rate_matrix(self) -> np.ndarray:
        return (1/np.cos(self.a2))*np.array([
            [
                0,
                np.sin(self.a3),
                np.cos(self.a3)
            ],
            [
                0,
                np.cos(self.a2)*np.cos(self.a3),
                -np.cos(self.a2)*np.sin(self.a3)
            ],
            [
                np.cos(self.a2),
                np.sin(self.a2)*np.sin(self.a3),
                np.sin(self.a2)*np.cos(self.a3)
            ]
        ])

    @classmethod
    def from_dcm(
        cls, dcm: "direction_cosine_matrix.DCM"
    ) -> "EulerAngle321":
        C = dcm.array
        a1 = np.arctan2(C[0,1], C[0,0])
        a2 = -np.arcsin(C[0,2])
        a3 = np.arctan2(C[1,2], C[2,2])
        return cls(a1, a2, a3)


class EulerAngle313(BaseSymmetricEulerAngle):
    """3-1-3 Euler Angle (yaw, roll, yaw)."""

    def from_dcm(
        cls, dcm: "direction_cosine_matrix.DCM"
    ) -> "EulerAngle313":
        C = dcm.array
        a1 = np.arctan2(C[2,0], -C[2,1])
        a2 = np.arcnp.cos(C[2,2])
        a3 = np.arctan2(C[0,2], C[1,2])
        return cls(a1, a2, a3)



"""Base classes for attitude parameterizations."""

import numpy as np


class BaseAttitudeParameter:
    """Base class for attitude parameterizations."""

    def __init__(self, array: np.ndarray) -> None:
        """
        Abstract initialization based on numpy array of values.
        Derived classes should define how to initialize based on
        these values.
        """



"""Implements Direction Cosine Matrix (DCM) attitude parameter."""

from typing import Optional

import numpy as np

import attlesnake.parameter.base as base
import attlesnake.parameter.euler_angle as euler_angle


class DCM(base.BaseAttitudeParameter):
    """Direction Cosine Matrix (DCM)."""
    def __init__(
        self, array_3x3: Optional[float] = None) -> None:
        if array_3x3 is None:
            self.array = np.eye(3)
        else:
            if not np.shape(array_3x3) == (3, 3):
                raise ValueError("DCM must be initialized with a 3x3 array of floats")
            # TODO: add validation that given array represents a valid DCM
            # (i.e. that columns are orthonormal with determinant +1)
            self.array = np.array(array_3x3)

    def __repr__(self):
        return (
            f"{round(self.array[0,0], 4)},\t{round(self.array[0,1], 4)},\t{round(self.array[0,2], 4)}\n"
            f"{round(self.array[1,0], 4)},\t{round(self.array[1,1], 4)},\t{round(self.array[1,2], 4)}\n"
            f"{round(self.array[2,0], 4)},\t{round(self.array[2,1], 4)},\t{round(self.array[2,2], 4)}"
        )

    @classmethod
    def from_ea321(
        cls, ea_321: "euler_angle.EulerAngle321"
    ) -> None:
        """
        Initialize the DCM from a 3-2-1 set of Euler
        angles, where the angles are expressed in radians.
        3-2-1: a1*yaw, a2*pitch, a3*roll.
        """
        a1, a2, a3 = ea_321.array
        array = np.array([
            [
                np.cos(a2)*np.cos(a1),
                np.cos(a2)*np.sin(a1),
                -np.sin(a2)
            ],
            [
                np.sin(a3)*np.sin(a2)*np.cos(a1) - np.cos(a3)*np.sin(a1),
                np.sin(a3)*np.sin(a2)*np.sin(a1) + np.cos(a3)*np.cos(a1),
                np.sin(a3)*np.cos(a2)
            ],
            [
                np.cos(a3)*np.sin(a2)*np.cos(a1) + np.sin(a3)*np.sin(a1),
                np.cos(a3)*np.sin(a2)*np.sin(a1) - np.sin(a3)*np.cos(a1),
                np.cos(a3)*np.cos(a2)
            ]
        ])
        return cls(array)

    @classmethod
    def from_ea313(
        cls, ea_313: "euler_angle.EulerAngle313"
    ) -> "DCM":
        """
        Initialize the DCM from a 3-1-3 set of Euler
        angles, where the angles are expressed in radians.
        3-1-3: a1*yaw, a2*roll, a3*yaw.
        """
        a1, a2, a3 = ea_313.array
        array = np.array([
            [
                np.cos(a3)*np.cos(a1) - np.sin(a3)*np.cos(a2)*np.sin(a1),
                np.cos(a3)*np.sin(a1) + np.sin(a3)*np.cos(a2)*np.cos(a1),
                np.sin(a3)*np.sin(a2)
            ],
            [
                -np.sin(a3)*np.cos(a1) - np.cos(a3)*np.cos(a2)*np.sin(a1),
                -np.sin(a3)*np.sin(a1) + np.cos(a3)*np.cos(a2)*np.cos(a1),
                np.cos(a3)*np.sin(a2)
            ],
            [
                np.sin(a2)*np.sin(a1),
                -np.sin(a2)*np.cos(a1),
                np.cos(a2)
            ]
        ])
        return cls(array)
